# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from threading import Event
from threading import Thread
import unittest

import launch
import launch.actions
import launch_ros.actions
import launch_testing.actions
from launch_testing.io_handler import ActiveIoHandler
import launch_testing.markers
import pytest
import rclpy
from std_msgs.msg import String


@pytest.mark.launch_test
@launch_testing.markers.keep_alive
def generate_test_description():
    path_to_test = os.path.dirname(__file__)

    return launch.LaunchDescription([
        launch_ros.actions.Node(
            executable=sys.executable,
            arguments=[os.path.join(path_to_test, 'talker.py')],
            additional_env={'PYTHONUNBUFFERED': '1'},
            name='demo_node_1',
        ),

        launch_testing.actions.ReadyToTest()
    ])


class TestFixture(unittest.TestCase):

    def subscription_callback(self, data: String):
        self.msg_event_object.set()

    def spin(self):
        try:
            while rclpy.ok() and not self.spinning.is_set():
                rclpy.spin_once(self.node, timeout_sec=0.1)
        finally:
            return

    def setUp(self):
        rclpy.init()
        self.node = rclpy.create_node('test_node')
        self.spinning = Event()
        self.msg_event_object = Event()
        self.subscription = self.node.create_subscription(
            String,
            'chatter',
            self.subscription_callback,
            10
        )

        # Add a spin thread
        self.ros_spin_thread = Thread(target=self.spin)
        self.ros_spin_thread.start()

    def tearDown(self):
        self.spinning.set()
        self.ros_spin_thread.join()
        self.node.destroy_subscription(self.subscription)
        self.node.destroy_node()
        rclpy.shutdown()

    def test_check_if_msgs_published(self, proc_output: ActiveIoHandler):
        msgs_received_flag = self.msg_event_object.wait(timeout=15.0)
        assert msgs_received_flag, 'Did not receive msgs !'



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Only import standard python modules here
# This module intentionally delays imports as late as possible to avoid
# importing downstream modules in upstream packages when built with a merged
# workspace.

import pytest


def pytest_launch_collect_makemodule(path, parent, entrypoint):
    marks = getattr(entrypoint, 'pytestmark', [])
    if marks and any(m.name == 'rostest' for m in marks):
        from launch_testing_ros.pytest.hooks import LaunchROSTestModule
        module = LaunchROSTestModule.from_parent(parent=parent, fspath=path)
        for mark in marks:
            decorator = getattr(pytest.mark, mark.name)
            decorator = decorator.with_args(*mark.args, **mark.kwargs)
            module.add_marker(decorator)
        return module


def pytest_configure(config):
    config.addinivalue_line(
        'markers',
        'rostest: mark a generate_test_description function as a ROS launch test entrypoint'
    )



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import unittest

import launch
import launch.actions
import launch_ros.actions
import launch_testing.actions
import launch_testing.markers
from launch_testing_ros import WaitForTopics

import pytest

from std_msgs.msg import String


def generate_node(i: int):
    """Return node and remap the topic based on the index provided."""
    path_to_test = os.path.dirname(__file__)
    return launch_ros.actions.Node(
        executable=sys.executable,
        arguments=[os.path.join(path_to_test, 'talker.py')],
        name='demo_node_' + str(i),
        additional_env={'PYTHONUNBUFFERED': '1'},
        parameters=[{'hz': 10.0}],
        remappings=[('chatter', 'chatter_' + str(i))]
    )


@pytest.mark.launch_test
@launch_testing.markers.keep_alive
def generate_test_description():
    # 'n' changes the number of nodes and topics generated for this test
    n = 5
    description = [generate_node(i) for i in range(n)] + [launch_testing.actions.ReadyToTest()]
    return launch.LaunchDescription(description), {'count': n}


# TODO: Test cases fail on Windows debug builds
# https://github.com/ros2/launch_ros/issues/292
if os.name != 'nt':
    class TestFixture(unittest.TestCase):

        def test_topics_successful(self, count: int):
            """All the supplied topics should be read successfully."""
            topic_list = [('chatter_' + str(i), String) for i in range(count)]
            expected_topics = {'chatter_' + str(i) for i in range(count)}
            message_pattern = re.compile(r'Hello World: \d+')

            # Method 1 : Using the magic methods and 'with' keyword
            with WaitForTopics(
                topic_list, timeout=15.0, messages_received_buffer_length=10
            ) as wait_for_node_object_1:
                assert wait_for_node_object_1.topics_received() == expected_topics
                assert wait_for_node_object_1.topics_not_received() == set()
                for topic_name, _ in topic_list:
                    assert len(wait_for_node_object_1.received_messages(topic_name)) >= 1
                    message = wait_for_node_object_1.received_messages(topic_name).pop().data
                    assert message_pattern.match(message)

            # Multiple instances of WaitForNode() can be created safely as
            # their internal nodes spin in separate contexts
            # Method 2 : Manually calling wait() and shutdown()
            wait_for_node_object_2 = WaitForTopics(topic_list, timeout=15.0)
            assert wait_for_node_object_2.wait()
            assert wait_for_node_object_2.topics_received() == expected_topics
            assert wait_for_node_object_2.topics_not_received() == set()
            for topic_name, _ in topic_list:
                assert len(wait_for_node_object_1.received_messages(topic_name)) >= 1
                message = wait_for_node_object_2.received_messages(topic_name).pop().data
                assert message_pattern.match(message)
            wait_for_node_object_2.shutdown()

        def test_topics_unsuccessful(self, count: int):
            """All topics should be read except for the 'invalid_topic'."""
            topic_list = [('chatter_' + str(i), String) for i in range(count)]
            # Add a topic that will never have anything published on it
            topic_list.append(('invalid_topic', String))
            expected_topics = {'chatter_' + str(i) for i in range(count)}

            # Method 1
            with pytest.raises(RuntimeError):
                with WaitForTopics(topic_list, timeout=2.0):
                    pass

            # Method 2
            wait_for_node_object = WaitForTopics(topic_list, timeout=2.0)
            assert not wait_for_node_object.wait()
            assert wait_for_node_object.topics_received() == expected_topics
            assert wait_for_node_object.topics_not_received() == {'invalid_topic'}
            wait_for_node_object.shutdown()



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time
import unittest

import launch
import launch.actions
import launch_ros.actions
import launch_testing.actions
from launch_testing.io_handler import ActiveIoHandler
import launch_testing.markers
import pytest
import rclpy


@pytest.mark.launch_test
@launch_testing.markers.keep_alive
def generate_test_description():
    path_to_test = os.path.dirname(__file__)

    return launch.LaunchDescription([
        launch.actions.TimerAction(
            period=5.0,
            actions=[
                launch_ros.actions.Node(
                    executable=sys.executable,
                    arguments=[os.path.join(path_to_test, 'talker.py')],
                    additional_env={'PYTHONUNBUFFERED': '1'},
                    name='demo_node_1',
                )
            ]),
        launch_testing.actions.ReadyToTest()
    ])


class TestFixture(unittest.TestCase):

    def setUp(self):
        rclpy.init()
        self.node = rclpy.create_node('test_node')

    def tearDown(self):
        self.node.destroy_node()
        rclpy.shutdown()

    def test_node_start(self, proc_output: ActiveIoHandler):
        found = False
        print('Waiting for node...')
        # demo_node_1 won't start for at least 5 seconds after this test
        # is launched, so we wait for a total of up to 20 seconds for it
        # to appear.
        start = time.time()
        while time.time() - start < 20.0 and not found:
            found = 'demo_node_1' in self.node.get_node_names()
            time.sleep(0.1)

        assert found, 'Node not found!'



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
import random
import string
from threading import Event
from threading import Thread

import rclpy
from rclpy.executors import SingleThreadedExecutor
from rclpy.node import Node


class WaitForTopics:
    """
    Wait to receive messages on supplied topics.

    Example usage:

    from std_msgs.msg import String

    .. code-block:: python

        # Method 1, using the 'with' keyword
        def method_1():
            topic_list = [('topic_1', String), ('topic_2', String)]
            with WaitForTopics(topic_list, timeout=5.0):
                # 'topic_1' and 'topic_2' received at least one message each
                print('Given topics are receiving messages !')

        # Method 2, calling wait() and shutdown() manually
        def method_2():
            topic_list = [('topic_1', String), ('topic_2', String)]
            wait_for_topics = WaitForTopics(topic_list, timeout=5.0)
            assert wait_for_topics.wait()
            print('Given topics are receiving messages !')
            print(wait_for_topics.topics_not_received()) # Should be an empty set
            print(wait_for_topics.topics_received()) # Should be {'topic_1', 'topic_2'}
            print(wait_for_topics.messages_received('topic_1')) # Should be [message_1, ...]
            wait_for_topics.shutdown()
    """

    def __init__(self, topic_tuples, timeout=5.0, messages_received_buffer_length=10):
        self.topic_tuples = topic_tuples
        self.timeout = timeout
        self.messages_received_buffer_length = messages_received_buffer_length
        self.__ros_context = rclpy.Context()
        rclpy.init(context=self.__ros_context)
        self.__ros_executor = SingleThreadedExecutor(context=self.__ros_context)

        self._prepare_ros_node()

        # Start spinning
        self.__ros_spin_thread = Thread(target=self._spin_handle_external_shutdown)
        self.__ros_spin_thread.start()

    def _spin_handle_external_shutdown(self):
        try:
            self.__ros_executor.spin()
        except rclpy.executors.ExternalShutdownException:
            pass

    def _prepare_ros_node(self):
        node_name = '_test_node_' + ''.join(
            random.choices(string.ascii_uppercase + string.digits, k=10)
        )
        self.__ros_node = _WaitForTopicsNode(
            name=node_name,
            node_context=self.__ros_context,
            messages_received_buffer_length=self.messages_received_buffer_length,
        )
        self.__ros_executor.add_node(self.__ros_node)

    def wait(self):
        self.__ros_node.start_subscribers(self.topic_tuples)
        return self.__ros_node.msg_event_object.wait(self.timeout)

    def shutdown(self):
        # Shutdown context before joining thread
        self.__ros_context.try_shutdown()
        self.__ros_spin_thread.join()
        self.__ros_node.destroy_node()

    def topics_received(self):
        """Topics that received at least one message."""
        return self.__ros_node.received_topics

    def topics_not_received(self):
        """Topics that did not receive any messages."""
        return self.__ros_node.expected_topics - self.__ros_node.received_topics

    def received_messages(self, topic_name):
        """List of received messages of a specific topic."""
        if topic_name not in self.__ros_node.received_messages_buffer:
            raise KeyError('No messages received for topic: ' + topic_name)
        return list(self.__ros_node.received_messages_buffer[topic_name])

    def __enter__(self):
        if not self.wait():
            raise RuntimeError('Did not receive messages on these topics: ',
                               self.topics_not_received())
        return self

    def __exit__(self, exep_type, exep_value, trace):
        self.shutdown()


class _WaitForTopicsNode(Node):
    """Internal node used for subscribing to a set of topics."""

    def __init__(
            self, name='test_node', node_context=None, messages_received_buffer_length=None
    ):
        super().__init__(node_name=name, context=node_context)  # type: ignore
        self.msg_event_object = Event()
        self.messages_received_buffer_length = messages_received_buffer_length
        self.subscriber_list = []
        self.topic_tuples = []
        self.expected_topics = set()
        self.received_topics = set()
        self.received_messages_buffer = {}

    def _reset(self):
        self.msg_event_object.clear()
        self.received_topics = set()
        for buffer in self.received_messages_buffer.values():
            buffer.clear()

    def start_subscribers(self, topic_tuples):
        self._reset()
        for topic_name, topic_type in topic_tuples:
            if (topic_name, topic_type) not in self.topic_tuples:
                self.topic_tuples.append((topic_name, topic_type))
                self.expected_topics.add(topic_name)
                # Initialize ring buffer of received messages
                self.received_messages_buffer[topic_name] = deque(
                    maxlen=self.messages_received_buffer_length
                )
                # Create a subscriber
                self.subscriber_list.append(
                    self.create_subscription(
                        topic_type,
                        topic_name,
                        self.callback_template(topic_name),
                        10
                    )
                )

    def callback_template(self, topic_name):
        def topic_callback(data):
            self.get_logger().debug('Message received for ' + topic_name)
            self.received_messages_buffer[topic_name].append(data)
            if topic_name not in self.received_topics:
                self.received_topics.add(topic_name)
            if self.received_topics == self.expected_topics:
                self.msg_event_object.set()

        return topic_callback



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from . import tools
from .data_republisher import DataRepublisher
from .message_pump import MessagePump
from .test_runner import LaunchTestRunner
from . wait_for_topics import WaitForTopics

__all__ = [
    'DataRepublisher',
    'LaunchTestRunner',
    'MessagePump',
    'tools',
    'WaitForTopics',
]



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import rclpy
from rclpy.node import Node

from std_msgs.msg import String


class Listener(Node):

    def __init__(self):
        super().__init__('listener')
        self.subscription = self.create_subscription(
            String, 'chatter', self.callback, 10
        )

    def callback(self, msg: String):
        self.get_logger().info('I heard: [%s]' % msg.data)


def main(args=None):
    rclpy.init(args=args)

    node = Listener()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    finally:
        node.destroy_node()
        rclpy.shutdown()


if __name__ == '__main__':
    main()



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DataRepublisher:
    """Republish mesasges with a transform function applied."""

    def __init__(self, node, listen_topic, publish_topic, msg_type, transform_fn):
        """
        Create a DataRepublisher.

        :param node: A rclpy node that will run the publisher and subscriber

        :param listen_topic: The topic to listen for incoming messages on

        :param publish_topic: The topic to republish messages on

        :param msg_type: The type of ROS msg to receive and republish

        :param transfor_fn: A function that takes one mesasge and returns a new message to
        republish or None to drop the message
        """
        self.__num_received = 0
        self.__republished_list = []

        self.__node = node
        self.__subscriber = node.create_subscription(
            msg_type,
            listen_topic,
            self.__cb,
            10
        )

        self.__publisher = node.create_publisher(
            msg_type,
            publish_topic,
            10
        )

        self.__transform_fn = transform_fn

    def shutdown(self):
        """Stop republishing messages."""
        self.__node.destroy_subscription(self.__subscriber)
        self.__node.destroy_publisher(self.__publisher)

    def get_num_received(self):
        """Get the number of messages received on the listen_topic."""
        return self.__num_received

    def get_num_republished(self):
        """
        Get the number of messages published on publish_topic.

        This may be lower than get_num_received if the transform_fn indicated a message
        should be dropped
        """
        return len(self.__republished_list)

    def get_republished(self):
        """Get a list of all of the transformed messages republished."""
        return self.__republished_list

    def __cb(self, msg):
        self.__num_received += 1
        repub = self.__transform_fn(msg)

        if repub:
            self.__republished_list.append(msg)
            self.__publisher.publish(repub)
